﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Immutable;

namespace Encompass
{
    /// <summary>
    /// EntitySetQuery is used to efficiently obtain a set of Entities that have all required Components and do not have any forbidden Components.
    /// </summary>
    public struct EntitySetQuery : IEnumerable<Entity>
    {
        private EntitySetQuery(EntityManager entityManager, ComponentUpdateManager componentUpdateManager, HashSet<Type> readTypes, HashSet<Type> readPendingTypes, ImmutableArray<Type> includes, ImmutableArray<Type> excludes)
        {
            EntityManager = entityManager;
            ComponentUpdateManager = componentUpdateManager;
            ReadTypes = readTypes;
            ReadPendingTypes = readPendingTypes;
            Includes = includes;
            Excludes = excludes;
        }

        internal EntitySetQuery(EntityManager entityManager, ComponentUpdateManager componentUpdateManager, HashSet<Type> readTypes, HashSet<Type> readPendingTypes)
        {
            EntityManager = entityManager;
            ComponentUpdateManager = componentUpdateManager;
            ReadTypes = readTypes;
            ReadPendingTypes = readPendingTypes;
            Includes = ImmutableArray.Create<Type>();
            Excludes = ImmutableArray.Create<Type>();
        }

        private EntityManager EntityManager { get; }
        private ComponentUpdateManager ComponentUpdateManager { get; }
        private HashSet<Type> ReadTypes { get; }
        private HashSet<Type> ReadPendingTypes { get; }
        private ImmutableArray<Type> Includes { get; }
        private ImmutableArray<Type> Excludes { get; }

        /// <summary>
        /// Designates that the given component type is required.
        /// </summary>
        public EntitySetQuery With<TComponent>() where TComponent : struct, IComponent
        {
            return new EntitySetQuery(EntityManager, ComponentUpdateManager, ReadTypes, ReadPendingTypes, Includes.Add(typeof(TComponent)), Excludes);
        }

        /// <summary>
        /// Designates that the given component type is forbidden.
        /// </summary>
        public EntitySetQuery Without<TComponent>() where TComponent : struct, IComponent
        {
            return new EntitySetQuery(EntityManager, ComponentUpdateManager, ReadTypes, ReadPendingTypes, Includes, Excludes.Add(typeof(TComponent)));
        }

        public IEnumerator<Entity> GetEnumerator()
        {
            return ComponentUpdateManager.QueryEntities(EntityManager.Entities, ReadTypes, ReadPendingTypes, Includes, Excludes).GetEnumerator();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return GetEnumerator();
        }
    }
}
